import typing
from tkinter import Canvas


class FontConfig(typing.NamedTuple):
    text: str
    size: int


class SidePage(typing.NamedTuple):
    name: str

    cav: typing.Type[Canvas]


class SidePageConfig(typing.NamedTuple):
    """
    Side page configuration.
    """
    width: int
    height: int
    side_pages: list[SidePage]
    side_button_height: int
    title: FontConfig
    subtitle: FontConfig


"""All containers.

There are two containers at the window level: `Tk` and `Toplevel`. `Tk` is
generally used for the main window, while `Toplevel` is a pop-up window.

There is another container at the canvas level: `Canvas`. `Canvas` is the main
container carrier.
"""

from __future__ import annotations

__all__ = [
    'PageCanvas',
]

import typing

from maliang.core.containers import Canvas, Tk, Toplevel
from .config import *

import maliang


class PageCanvas(Canvas):
    def __init__(
            self,
            side_config: SidePageConfig,
            master: Tk | Toplevel | Canvas | None = None,
            *,
            expand: typing.Literal["", "x", "y", "xy"] = "xy",
            auto_zoom: bool = False,
            default: int = 0,
            keep_ratio: typing.Literal["min", "max"] | None = None,
            free_anchor: bool = False,
            auto_update: bool | None = None,
            zoom_all_items: bool = False,
            **kwargs,
    ) -> None:
        """
        * `master`: parent widget
        * `expand`: the mode of expand, `x` is horizontal, and `y` is vertical
        * `auto_zoom`: whether or not to scale its items automatically
        * `keep_ratio`: the mode of aspect ratio, `min` follows the minimum
        value, `max` follows the maximum value
        * `free_anchor`: whether the anchor point is free-floating
        * `auto_update`: whether the theme manager update it automatically
        * `zoom_all_items`: (Experimental) whether or not to scale its all items
        * `kwargs`: compatible with other parameters of class `tkinter.Canvas`
        """
        super().__init__(
            master,
            expand=expand,
            auto_zoom=auto_zoom,
            keep_ratio=keep_ratio,
            free_anchor=free_anchor,
            auto_update=auto_update,
            zoom_all_items=zoom_all_items,
            **kwargs
        )
        frame_side = maliang.Canvas(self, width=side_config.width, expand="x")
        frame_side.pack(fill="y", side="left")
        self.frame_main = frame_main = maliang.Canvas(self, auto_zoom=True)
        frame_main.pack(fill="both", expand=True)

        self.side_pages = side_config.side_pages

        self.setup_side(frame_side, side_config, default)

    def setup_side(self, frame_side: maliang.Canvas, side_config: SidePageConfig, default: int) -> None:
        canvas = maliang.Canvas(
            frame_side, expand="y", highlightthickness=0)
        canvas.place(width=side_config.width, height=side_config.height)
        maliang.Text(canvas, (side_config.width // 2, side_config.side_button_height),
                             text=side_config.title.text,
                             fontsize=side_config.title.size, anchor="center")
        maliang.Text(
            canvas,
            (side_config.width // 2, int(side_config.side_button_height * 1.8)),
            text=side_config.subtitle.text,
            fontsize=side_config.subtitle.size, anchor="center")

        sizes = ((side_config.width - 50, 50),) * len(side_config.side_pages)
        text = []

        for page in side_config.side_pages:
            text.append(page.name)

        maliang.SegmentedButton(canvas, (20, 140), sizes,
                                default=default, text=tuple(text), command=self.call_canvas, layout="vertical")

    def call_canvas(self, index: int) -> None:
        """"""
        self.side_pages[index].cav(self.frame_main, auto_zoom=True, free_anchor=True, keep_ratio="min")



import math

import maliang
import maliang.animation as animation


class LoginToplevel(maliang.Toplevel):
    """登录窗口"""

    def load_ui(self) -> None:
        """加载 UI"""
        canvas = maliang.Canvas(self)
        canvas.place(width=480, height=720)

        self.sub_title = maliang.Text(
            canvas, (240, 45), text="登录到你的账号", fontsize=36, anchor="center")

        canvas.create_oval(120, 90, 360, 330, outline="grey")
        canvas.create_text(240, 210, text="用户\n头像", fill="grey", font=30)

        self.account = maliang.InputBox(
            canvas, (40, 360), (400, 50), placeholder="请输入您的账号")
        self.password = maliang.InputBox(
            canvas, (40, 430), (400, 50), placeholder="请输入您的密码", show="●")
        self.an = maliang.Button(canvas, (40, 500), (190, 50),
                                 text="注 册", command=self.animate)
        self.login = maliang.Button(canvas, (250, 500), (190, 50), text="登 录")
        self.password_verify = maliang.InputBox(
            canvas, (40, 500+300), (400, 50), placeholder="请再次输入您的密码", show="●")
        self.registry = maliang.Button(
            canvas, (40-300, 570), (190, 50), text="注 册")
        self.back = maliang.Button(canvas, (250+300, 570),
                                   (190, 50), text="返 回", command=lambda: self.animate(True))

        self.forget = maliang.UnderlineButton(
            canvas, (140, 600), text="忘记密码", fontsize=20, anchor="center")
        self.sep = maliang.Text(canvas, (190, 600), text="|", anchor="center")
        self.find = maliang.UnderlineButton(
            canvas, (240, 600), text="找回账号", fontsize=20, anchor="center")
        self.sep_2 = maliang.Text(canvas, (290, 600), text="|", anchor="center")
        self.net = maliang.UnderlineButton(
            canvas, (340, 600), text="网络设置", fontsize=20, anchor="center")
        self.animation_lock = False  # 防熊

    def animate(self, back: bool = False) -> None:
        """执行相关动画"""
        if self.animation_lock:
            return
        self.animation_lock = True
        k = -1 if back else 1
        self.after(
            250, self.sub_title.texts[0].set, "登录到你的账号" if back else "注册新的账号")
        self.after(
            250, self.title, "登录" if back else "注册")
        animation.MoveWidget(self.sub_title, (0, -80), 500,
                             controller=animation.generate(math.sin, 0, math.pi, map_y=False), fps=60,
                             end=lambda: self.__setattr__("animation_lock", False)).start()
        animation.MoveWidget(self.an, (-300*k, 0), 500,
                             controller=animation.smooth, fps=60).start()
        animation.MoveWidget(self.login, (300*k, 0), 500,
                             controller=animation.smooth, fps=60).start()
        animation.MoveWidget(self.registry, (300*k, 0), 500,
                             controller=animation.smooth, fps=60).start()
        animation.MoveWidget(self.back, (-300*k, 0), 500,
                             controller=animation.smooth, fps=60).start()
        animation.MoveWidget(self.forget, (0, 100*k), 500,
                             controller=animation.smooth, fps=60).start()
        animation.MoveWidget(self.sep, (0, 100*k), 500,
                             controller=animation.smooth, fps=60).start()
        animation.MoveWidget(self.find, (0, 100*k), 500,
                             controller=animation.smooth, fps=60).start()
        animation.MoveWidget(self.sep_2, (0, 100*k), 500,
                             controller=animation.smooth, fps=60).start()
        animation.MoveWidget(self.net, (0, 100*k), 500,
                             controller=animation.smooth, fps=60).start()
        animation.MoveWidget(self.password_verify, (0, -300*k), 500,
                             controller=animation.smooth, fps=60).start()


"""All standard `Feature` classes"""

from __future__ import annotations

__all__ = [
    "IconOnlyFeature",
    "Underline",
    "Highlight",
]

import tkinter

from maliang.animation import animations, controllers
from maliang.toolbox import utility
from maliang.standard.features import ButtonFeature


class IconOnlyFeature(ButtonFeature):
    """Feature of Icon Only Button"""

    def _motion(self, event: tkinter.Event) -> bool:
        if flag := self.widget.images[0].detect(event.x, event.y):
            cursor = utility.fix_cursor(
                "disabled" if self.widget.state == "disabled" else "hand2")
            self.widget.master.trigger_config.update(cursor=cursor)
            if self.widget.state == "normal":
                self.widget.update("hover")
        else:
            if self.widget.state != "normal":
                self.widget.update("normal")
        return flag

    def _button_release_1(self, event: tkinter.Event) -> bool:
        if flag := self.widget.images[0].detect(event.x, event.y):
            if self.widget.state == "active":
                self.widget.update("hover")
                if self.command is not None:
                    self.command(*self._args)
        return flag


class Underline(ButtonFeature):
    """Feature of underline"""

    def _motion(self, event: tkinter.Event) -> bool:
        if flag := self.widget.texts[0].detect(event.x, event.y):
            cursor = utility.fix_cursor(
                "disabled" if self.widget.state == "disabled" else "hand2")
            self.widget.master.trigger_config.update(cursor=cursor)
            if self.widget.state == "normal":
                self.widget.update("hover")
                self.widget.texts[0].font.config(underline=True)
        else:
            if self.widget.state != "normal":
                self.widget.update("normal")
                self.widget.texts[0].font.config(underline=False)
        return flag

    def _button_1(self, _: tkinter.Event) -> bool:
        if flag := self.widget.state == "hover":
            self.widget.update("active")
        return flag

    def _button_release_1(self, event: tkinter.Event) -> bool:
        if flag := self.widget.texts[0].detect(event.x, event.y):
            if self.widget.state == "active":
                self.widget.update("hover")
                self.widget.texts[0].font.config(underline=True)
                if self.command is not None:
                    self.command(*self._args)
        return flag


class Highlight(ButtonFeature):
    """Feature of highlight"""

    def _motion(self, event: tkinter.Event) -> bool:
        if flag := self.widget.texts[0].detect(event.x, event.y):
            cursor = utility.fix_cursor(
                "disabled" if self.widget.state == "disabled" else "hand2")
            self.widget.master.trigger_config.update(cursor=cursor)
            if self.widget.state == "normal":
                self.widget.update("hover")
                animations.ScaleFontSize(self.widget.texts[0], 28, 150).start()
        else:
            if self.widget.state != "normal":
                self.widget.update("normal")
                animations.ScaleFontSize(self.widget.texts[0], 24, 150).start()
        return flag

    def _button_1(self, _: tkinter.Event) -> bool:
        if flag := self.widget.state == "hover":
            self.widget.update("active")
            animations.ScaleFontSize(self.widget.texts[0], 26, 150).start()
        return flag

    def _button_release_1(self, event: tkinter.Event) -> bool:
        if flag := self.widget.texts[0].detect(event.x, event.y):
            if self.widget.state == "active":
                self.widget.update("hover")
                animations.ScaleFontSize(self.widget.texts[0], 28, 150).start()
                if self.command is not None:
                    self.command(*self._args)
        return flag



"""All extensions `Widget` classes"""

from __future__ import annotations

__all__ = [
    'UnderlineButton',
    'HighlightButton',
    'IconButton',
    'IconOnlyButton',
]

import collections.abc
import typing

from maliang.core import configs, containers, virtual
from maliang.toolbox import enhanced, utility
from maliang.standard import images, shapes, styles, texts

from . import features


class UnderlineButton(virtual.Widget):
    """Underline button, generally used to display web links"""

    def __init__(
            self,
            master: containers.Canvas | virtual.Widget,
            position: tuple[int, int],
            *,
            text: str = "",
            family: str | None = None,
            fontsize: int | None = None,
            weight: typing.Literal['normal', 'bold'] = "normal",
            slant: typing.Literal['roman', 'italic'] = "roman",
            underline: bool = False,
            overstrike: bool = False,
            justify: typing.Literal["left", "center", "right"] = "left",
            command: collections.abc.Callable | None = None,
            image: enhanced.PhotoImage | None = None,
            anchor: typing.Literal["n", "e", "w", "s", "nw", "ne", "sw", "se", "center"] = "nw",
            capture_events: bool | None = None,
            gradient_animation: bool = False,
            auto_update: bool | None = None,
            style: type[virtual.Style] | None = None,
    ) -> None:
        """
        * `master`: parent canvas
        * `position`: position of the widget
        * `text`: text of the widget
        * `family`: font family
        * `fontsize`: font size
        * `weight`: weight of the text
        * `slant`: slant of the text
        * `underline`: whether the text is underline
        * `overstrike`: whether the text is overstrike
        * `justify`: justify mode of the text
        * `command`: a function that is triggered when the underline button is pressed
        * `image`: image of the widget
        * `anchor`: anchor of the widget
        * `capture_events`: wether detect another widget under the widget
        * `gradient_animation`: wether enable gradient_animation
        * `auto_update`: whether the theme manager update it automatically
        * `style`: style of the widget
        """
        virtual.Widget.__init__(
            self, master, position, utility.get_text_size(
                text, fontsize, family, weight=weight, slant=slant, master=master),
            anchor=anchor, capture_events=capture_events,
            gradient_animation=gradient_animation, auto_update=auto_update, style=style)
        if style is None:
            self.style = styles.UnderlineButtonStyle(self)
        if image is not None:
            images.StillImage(self, image=image)
        texts.Information(
            self, text=text, family=family, fontsize=fontsize, weight=weight, slant=slant,
            underline=underline, overstrike=overstrike, justify=justify)
        self.feature = features.Underline(self, command=command)


class HighlightButton(virtual.Widget):
    """Highlight button, no outline, which added a highlight effect"""

    def __init__(
            self,
            master: containers.Canvas | virtual.Widget,
            position: tuple[int, int],
            *,
            text: str = "",
            family: str | None = None,
            fontsize: int | None = None,
            weight: typing.Literal['normal', 'bold'] = "normal",
            slant: typing.Literal['roman', 'italic'] = "roman",
            underline: bool = False,
            overstrike: bool = False,
            justify: typing.Literal["left", "center", "right"] = "left",
            command: collections.abc.Callable | None = None,
            image: enhanced.PhotoImage | None = None,
            anchor: typing.Literal["n", "e", "w", "s", "nw", "ne", "sw", "se", "center"] = "nw",
            capture_events: bool | None = None,
            gradient_animation: bool | None = None,
            auto_update: bool | None = None,
            style: type[virtual.Style] | None = None,
    ) -> None:
        """
        * `master`: parent canvas
        * `position`: position of the widget
        * `text`: text of the widget
        * `family`: font family
        * `fontsize`: font size
        * `weight`: weight of the text
        * `slant`: slant of the text
        * `underline`: whether the text is underline
        * `overstrike`: whether the text is overstrike
        * `justify`: justify mode of the text
        * `command`: a function that is triggered when the hightlight button is pressed
        * `image`: image of the widget
        * `anchor`: anchor of the widget
        * `capture_events`: wether detect another widget under the widget
        * `gradient_animation`: wether enable gradient_animation
        * `auto_update`: whether the theme manager update it automatically
        * `style`: style of the widget
        """
        virtual.Widget.__init__(
            self, master, position, utility.get_text_size(
                text, fontsize, family, weight=weight, slant=slant, master=master),
            anchor=anchor, capture_events=capture_events,
            gradient_animation=gradient_animation, auto_update=auto_update, style=style)
        if style is None:
            self.style = styles.HighlightButtonStyle(self)
        if image is not None:
            images.StillImage(self, image=image)
        texts.Information(
            self, text=text, family=family, fontsize=fontsize, weight=weight, slant=slant,
            underline=underline, overstrike=overstrike, justify=justify)
        self.feature = features.Highlight(self, command=command)


class IconButton(virtual.Widget):
    """A button with an icon on the left side"""

    def __init__(
            self,
            master: containers.Canvas | virtual.Widget,
            position: tuple[int, int],
            size: tuple[int, int] | None = None,
            *,
            text: str = "",
            family: str | None = None,
            fontsize: int | None = None,
            weight: typing.Literal['normal', 'bold'] = "normal",
            slant: typing.Literal['roman', 'italic'] = "roman",
            underline: bool = False,
            overstrike: bool = False,
            justify: typing.Literal["left", "center", "right"] = "left",
            command: collections.abc.Callable | None = None,
            image: enhanced.PhotoImage | None = None,
            anchor: typing.Literal["n", "e", "w", "s", "nw", "ne", "sw", "se", "center"] = "nw",
            capture_events: bool | None = None,
            gradient_animation: bool | None = None,
            auto_update: bool | None = None,
            style: type[virtual.Style] | None = None,
    ) -> None:
        """
        * `master`: parent canvas
        * `position`: position of the widget
        * `size`: size of the widget
        * `text`: text of the widget
        * `family`: font family
        * `fontsize`: font size
        * `weight`: weight of the text
        * `slant`: slant of the text
        * `underline`: whether the text is underline
        * `overstrike`: whether the text is overstrike
        * `justify`: justify mode of the text
        * `command`: a function that is triggered when the button is pressed
        * `image`: image of the widget
        * `anchor`: anchor of the widget
        * `capture_events`: wether detect another widget under the widget
        * `gradient_animation`: wether enable gradient_animation
        * `auto_update`: whether the theme manager update it automatically
        * `style`: style of the widget
        """
        if size is None:
            size = utility.get_text_size(
                text, fontsize, family, weight=weight, slant=slant, padding=6, master=master)
            size = size[0] + size[1] - 6, size[1]
        virtual.Widget.__init__(
            self, master, position, size, anchor=anchor,
            capture_events=capture_events, gradient_animation=gradient_animation,
            auto_update=auto_update, style=style)
        if style is None:
            self.style = styles.IconButtonStyle(self)
        if configs.Env.system == "Windows10":
            shapes.Rectangle(self)
        else:
            shapes.RoundedRectangle(self)
        if image is not None:
            images.StillImage(self, ((size[1] - size[0]) / 2, 0), image=image)
        texts.Information(
            self, (size[1] - size[0] / 2, 0), text=text, family=family, fontsize=fontsize,
            weight=weight, slant=slant, underline=underline, overstrike=overstrike,
            justify=justify, anchor="w")
        self.feature = features.ButtonFeature(self, command=command)

    def get(self) -> str:
        """Get the text of the widget"""
        return self.texts[0].get()

    def set(self, text: str) -> None:
        """Set the text of the widget"""
        return self.texts[0].set(text)


class IconOnlyButton(IconButton):
    """A button with nothing but an icon"""

    def __init__(
            self,
            master: containers.Canvas | virtual.Widget,
            position: tuple[int, int],
            size: tuple[int, int] | None = None,
            *,
            command: collections.abc.Callable | None = None,
            image: enhanced.PhotoImage | None = None,
            borderless=True,
            anchor: typing.Literal["n", "e", "w", "s", "nw", "ne", "sw", "se", "center"] = "nw",
            capture_events: bool | None = None,
            gradient_animation: bool | None = None,
            auto_update: bool | None = None,
            style: type[virtual.Style] | None = None,
    ) -> None:
        """
        * `master`: parent canvas
        * `position`: position of the widget
        * `size`: size of the widget
        * `command`: a function that is triggered when the button is pressed
        * `image`: image of the widget
        * `anchor`: anchor of the widget
        * `capture_events`: wether detect another widget under the widget
        * `gradient_animation`: wether enable gradient_animation
        * `auto_update`: whether the theme manager update it automatically
        * `style`: style of the widget
        """
        if size is None:
            size = image.width(), image.height()
        virtual.Widget.__init__(
            self, master, position, size, anchor=anchor,
            capture_events=capture_events, gradient_animation=gradient_animation,
            auto_update=auto_update, style=style)
        if style is None:
            self.style = styles.TextStyle(self) if borderless else styles.IconButtonStyle(self)
        if image is not None:
            images.StillImage(self, ((size[1] - size[0]) / 2, 0), image=image)
        self.feature = features.IconOnlyFeature(self, command=command)

    def get(self) -> Exception:
        """This class has nothing to get"""
        raise AttributeError()

    def set(self, thing: None) -> Exception:
        """This class has nothing to set"""
        raise AttributeError()
